window.VoxelMesh = function VoxelMesh()
{
	var verts;
	var uvs;
	var normals;
	var colors;
	
	var sprite;
	var palette;
	var mx, my, mz;
	var sx, sy, sz;
	var ox=0, oy=0, oz=0;
	var gx=0, gy=0, gz=0;
	var cloud;

	var geometry = new THREE.BufferGeometry();
	var verts = new Array();
	var normals = new Array();
	var colors = new Array();
	var uvs = new Array();
	var bone = null;
	var skinInd = new Array();
	var skinW = new Array();
	var bones = null;
	var animations = new Array();
	
	function __ctor( data )
	{
		if( data.sprite )
		{
			sprite = data.sprite;
			sprite.margin = sprite.margin || 0.001;
		}
		else if( data.palette )
		{
			palette = data.palette;
			for( var i in palette ) palette[i] = new THREE.Color(palette[i]);
		}
		
		cloud = new Array();
		bone = data.root || null;
		var r = new Object();
		if( bone ) 
		{
			bones = new Array();
			r.id = data.root;
			r.pos = new Object();
			r.pos.x = r.pos.y = r.pos.z = 0;
			loadChild( data, r, -1 );
			
			anim = new Object();
			anim.name = "reset";
			anim.fps = 24;
			anim.length = 1;
			anim.hierarchy = new Array();
			animations.push(anim);
			
			loadAnimations( data, r );
		}
		else
		{
			r.id = "root";
			data.root = data;
			loadChild( data, r, -1 );
			delete data.root;
		}
	}
	
	
	var anim;
	function loadAnimations( scene, r )
	{
		var hi = anim.hierarchy[r.bone] || new Object();
		hi.parent = r.parent;
		hi.keys = hi.keys || new Array();
		
		var key = new Object();
		key.time = 0;
		key.scl = new Array(1,1,1);
		key.pos = new Array( -r.pos.x*key.scl[0], -r.pos.y*key.scl[1], -r.pos.z*key.scl[2] );
		key.rot = new Array(1, 0, 0, 0);
		hi.keys.push(key);
		
		anim.hierarchy[r.bone] = hi;
	
		if( scene[r.id].children )
		for( var i=0, c; c=scene[r.id].children[i]; ++i )
			loadAnimations( scene, c );

	}
	
	
	function loadChild( scene, ref, parentBone )
	{
		if( !ref ) return;
		var obj = scene[ref.id];
		if( !obj ) return;
		
		mx = obj.size[0];
		my = obj.size[1];
		mz = obj.size[2];

		if( !ref.pos ) ref.pos = new THREE.Vector3(0,0,0);
		if( bones )
		{
			var b = new Object();
			b.parent = parentBone;
			b.name = obj.name;
			b.rotq = new Array(1,0,0,0);
			b.scl = ref.scale || new Array(1,1,1);
			b.pos = new Array(
				/* * /
				0, 0, 0
				/*/
				-ref.pos.x*b.scl[0],
				-ref.pos.y*b.scl[1],
				-ref.pos.z*b.scl[2]
				/* */
				);
			bone = parentBone = bones.length;
			ref.bone = bone;
			ref.parent = parentBone;
			bones.push(b);
			sx=sy=sz=1;
		}
		sx = obj.scale[0] || 1;
		sy = obj.scale[1] || 1;
		sz = obj.scale[2] || 1;
		ox = (/**/ ref.pos.x + /**/obj.position[0])*sx || 0;
		oy = (/**/ ref.pos.y + /**/obj.position[1])*sy || 0;
		oz = (/**/ ref.pos.z + /**/obj.position[2])*sz || 0;
		
		if( !obj.uncompressed )
		{
			obj.points = uncompress( obj.points );
			obj.uncompressed = true;
		}
		cloud = obj.points;
		processCloud();
		
		if( obj.children )
		for( var i=0, c; c=obj.children[i]; ++i )
			loadChild( scene, c, parentBone );
	}
	
	function uncompress( m )
	{
		var out = new Array();
		for( var i=0; i<m.length; ++i )
		{
			var c = m[i];
			var j;
			if( c<0 )
			{
				j = c;
				c = m[++i];
				while(j++) out.push(c);
			}else out.push(c);
		}
		return out;
	}
	
	function processCloud()
	{
		
		var mzmx = mz*mx;
		for( var y=0; y<my; y++ )
		{
			for( var z=0; z<mz; z++ )
			{
				for( var x=0; x<mx; x++ )
				{
					var p = y*mzmx + z*mx + x;
					if( cloud[p] ) 
						addCube(
							x, y, z, cloud[p],
							y+1<my && cloud[p+mzmx],
							y>0 && cloud[p-mzmx],
							x+1<mx && cloud[p+1],
							x>0 && cloud[p-1],
							z+1<mz && cloud[p+mx],
							z>0 && cloud[p-mx]
						);
				}
			}
		}
		

	}
	
	function addCube( x, y, z, colorId, u, d, l, r, f, b )
	{
		var si = null;
		var c = null;
		var t = null;
		var lsx=1,lsy=1,lsz=1;
		
		if( typeof colorId == "object" )
		{
			if( colorId.scale )
			{
				lsx = colorId.scale[0];
				lsy = colorId.scale[1];
				lsz = colorId.scale[2];
			}
			if( colorId.offset )
			{
				x += colorId.offset[0];
				y += colorId.offset[1];
				z += colorId.offset[2];
			}
			if( palette ) colorId = colorId.color;
			else if( sprite ) colorId = colorId.tile;
		}
		if( lsx != 1 ) l = r = 0;
		if( lsz != 1 ) f = b = 0;
		if( lsy != 1 ) f = b = l = r = u = d = 0;
		
		x    = x*sx + ox;
		y    = y*sy + oy;
		z    = z*sz + oz;
		lsx *= sx;
		lsy *= sy;
		lsz *= sz;

		
		if( u && u.scale ) u = 0;
		if( d && d.scale ) d = 0;
		if( l && l.scale ) l = 0;
		if( r && r.scale ) r = 0;
		if( f && f.scale ) f = 0;
		if( b && b.scale ) b = 0;
		
		if( bone !== null )
		{
			si = new Array(
				bone, bone, bone,
				bone, bone, bone
				);
		}
		
		if( palette )
		{
			var color = palette[colorId];
			c = new Array(
					color.r, color.g, color.b, 
					color.r, color.g, color.b, 
					color.r, color.g, color.b, 
					color.r, color.g, color.b, 
					color.r, color.g, color.b, 
					color.r, color.g, color.b
					);
		}
		if( sprite )
		{
			var x0 = (colorId-1) % sprite.tileX;
			var x1 = x0 + 1;
			var y0 = Math.floor((colorId-1) / sprite.tileX);
			var y1 = y0 + 1;
			x0 /= sprite.tileX;
			x1 /= sprite.tileX;
			y0 /= sprite.tileX;
			y1 /= sprite.tileX;
			y0 = 1-(y0+sprite.margin);
			y1 = 1-(y1-sprite.margin);
			x0 += sprite.margin;
			x1 -= sprite.margin;
			t = new Array(
				x0, y0,
				x1, y0,
				x0, y1,
				x1, y1,
				x0, y1,
				x1, y0
			);
		}
				
		if( !u )
		{
			verts.push(
				x, y+lsy, z, 
				x, y+lsy, z+lsz,
				x+lsx, y+lsy, z, 
				
				x+lsx, y+lsy, z+lsz,
				x+lsx, y+lsy, z,
				x, y+lsy, z+lsz
				);

			if(si) skinInd.push.apply( skinInd, si );
			if(c) colors.push.apply( colors, c );
			else if(t) uvs.push.apply( uvs, t );
			normals.push(
				0,1,0,
				0,1,0,
				0,1,0,
				0,1,0,
				0,1,0,
				0,1,0
			);
		}
		
		if( !d )
		{
			verts.push(
				x, y, z, 
				x+lsx, y, z, 
				x, y, z+lsz,
				
				x+lsx, y, z+lsz,
				x, y, z+lsz,
				x+lsx, y, z
				);
			if(si) skinInd.push.apply( skinInd, si );
			if(c) colors.push.apply( colors, c );
			else if(t) uvs.push.apply( uvs, t );
			normals.push(
				0,-1,0,
				0,-1,0,
				0,-1,0,
				0,-1,0,
				0,-1,0,
				0,-1,0
			);
		}
		
		if( !r )
		{
			verts.push(
				x, y, z, 
				x, y, z+lsz,
				x, y+lsy, z, 
				
				x, y+lsy, z+lsz,
				x, y+lsy, z,
				x, y, z+lsz
				);
			if(si) skinInd.push.apply( skinInd, si );
			if(c) colors.push.apply( colors, c );
			else if(t) uvs.push.apply( uvs, t );
			normals.push(
				1,0,0,
				1,0,0,
				1,0,0,
				1,0,0,
				1,0,0,
				1,0,0
			);
		}
		
		if( !l )
		{
			verts.push(
				x+lsx, y, z, 
				x+lsx, y+lsy, z, 
				x+lsx, y, z+lsz,
				
				x+lsx, y+lsy, z+lsz,
				x+lsx, y, z+lsz,
				x+lsx, y+lsy, z 
				);
			if(si) skinInd.push.apply( skinInd, si );
			if(c) colors.push.apply( colors, c );
			else if(t) uvs.push.apply( uvs, t );
			normals.push(
				-1,0,0,
				-1,0,0,
				-1,0,0,
				-1,0,0,
				-1,0,0,
				-1,0,0
			);
		}
		
		if( !b )
		{
			verts.push(
				x, y, z, 
				x, y+lsy, z, 
				x+lsx, y, z,
				
				x+lsx, y+lsy, z,
				x+lsx, y, z,
				x, y+lsy, z
				);
			if(si) skinInd.push.apply( skinInd, si );
			if(c) colors.push.apply( colors, c );
			else if(t) uvs.push.apply( uvs, t );
			normals.push(
				0,0,-1,
				0,0,-1,
				0,0,-1,
				0,0,-1,
				0,0,-1,
				0,0,-1
			);
		}

		if( !f )
		{
			verts.push(
				x, y, z+lsz, 
				x+lsx, y, z+lsz,
				x, y+lsy, z+lsz, 
				
				x+lsx, y+lsy, z+lsz,
				x, y+lsy, z+lsz, 
				x+lsx, y, z+lsz
				);
			if(si) skinInd.push.apply( skinInd, si );
			if(c) colors.push.apply( colors, c );
			else if(t) uvs.push.apply( uvs, t );
			normals.push(
				0,0,1,
				0,0,1,
				0,0,1,
				0,0,1,
				0,0,1,
				0,0,1
			);
		}
	}

	this.getMesh = function getMesh()
	{
		var va = new Float32Array( verts.length );
		var na = new Float32Array( normals.length );
		
		for( var i=0; i<verts.length; ++i )
			va[i] = verts[i];

		for( var i=0; i<normals.length; ++i ) 
			na[i] = normals[i];
		
		if( colors.length )
		{
			var ca = new Float32Array( colors.length );
			for( var i=0; i<colors.length; ++i ) 
				ca[i] = colors[i];
			geometry.addAttribute( 'color', new THREE.BufferAttribute( ca, 3 ) );
		}
		
		if( uvs.length )
		{
			var ta = new Float32Array( uvs.length );
			for( var i=0; i<uvs.length; ++i ) 
				ta[i] = uvs[i];
			geometry.addAttribute( 'uv', new THREE.BufferAttribute( ta, 2 ) );
		}
		
		geometry.addAttribute( 'position', new THREE.BufferAttribute( va, 3 ) );
		geometry.addAttribute( 'normal', new THREE.BufferAttribute( na, 3 ) );
		
		if( skinInd.length )
		{
			var si = new Float32Array( skinInd.length * 4 );
			var sw = new Float32Array( skinInd.length * 4 );
			for( var i=0, j=0; i<skinInd.length; i++ )
			{
				sw[j] = 1; si[j++]=skinInd[i];
				sw[j] = 0; si[j++]=0;
				sw[j] = 0; si[j++]=0;
				sw[j] = 0; si[j++]=0;
			}
			
			geometry.addAttribute( 'skinIndex', new THREE.BufferAttribute( si, 4 ) );
			geometry.addAttribute( 'skinWeight', new THREE.BufferAttribute( sw, 4 ) );
			geometry.bones = bones;
			geometry.animations = animations;
			geometry.animation = animations[0];
		}
		
		var mesh = new Object();
		mesh.geometry = geometry;
		var mat = new Object();
		mat.ambient = 0xaaaaaa;
		mat.skinning = skinInd.length != 0;
		if( sprite && sprite.diffuse ) mat.map = THREE.ImageUtils.loadTexture( sprite.diffuse );
		else mat.vertexColors = THREE.VertexColors;
		mesh.material = new THREE.MeshPhongMaterial( mat );
		
		if( !skinInd.length )
			mesh.nodeClass = THREE.Mesh;
		else
			mesh.nodeClass = THREE.SkinnedMesh;
		
		mesh.create = function()
		{
			var tnode = new mesh.nodeClass( mesh.geometry, mesh.material );
			if( tnode.type == "SkinnedMesh" )
			{
				tnode.bone = new Object();
				for( var i=0, c; c=tnode.skeleton.bones[i]; ++i )
				{
					if( tnode.bone[ c.name ] )
					{
						var baseName = c.name;
						if( tnode.bone[ baseName ].name == baseName )
						{
							tnode.bone[ baseName ].name = baseName + "1";
							tnode.bone[ baseName + "1" ] = tnode.bone[ baseName ];
							tnode.bone[ c.name = baseName + "2" ] = c;
						}
						else
						{
							var n = 3;
							while( tnode.bone[ baseName + 3 ] ) n++;
							tnode.bone[ c.name = baseName + n ] = c;
						}
					}
					else
						tnode.bone[ c.name ] = c;
				}
			}
			return tnode;
		};
		
		return mesh;	
	}
	
	__ctor.apply( this, Array.prototype.slice.call(arguments) );
}
